#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   frame_fixture.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Upload Session Test Data
"""

# pylint: disable=C0301
register_frame_response = {
    "frame_format": "rectangle",
    "frame_id": "2fa565d3-27e2-4ee6-8362-bb4a3d8e7c12",
    "image_format": "pillow",
    "num_classes": 5,
    "project_id": "b705a30ae26671657f1fd51eb2d4739d",
    "shuffle": True,
    "train_count": 9,
    "validation_count": 419,
    "batch_size": 10,
}
